function  [uroot, planinfo, plandata] = osp(xroot, model, cfg)
% OCPD Optimistic Constrained planning for deterministic systems
%v2: added n and nsim option for cfg options. nsim=treesize, n=treesize/M

% Author: Lucian Busoniu & Mathe Koppany
% Versions:

CFG.gamma = [];                 % discount factor
CFG.n = [];                     % expansion budget
CFG.nsim = [];                     % expansion budget in terms of nr. of simulations
CFG.U = [];                     % discrete, flat action space
CFG.k = -1;                     % discrete time step
CFG.S = intmax;                 % nr. of allowed switches in action sequenvce
% display config
CFG.verb = 0;                   % verbosity: the higher, the more detailed the messages displayed

cfg = checkparams(cfg, CFG);

% quick-access to config parameters
% (note to avoid conflict with 'gamma' function, we must name discount factor differently)
discount = cfg.gamma; n = cfg.n; nsim = cfg.nsim;
U = cfg.U;
M = size(U, 2);     % number of discrete actions
S = cfg.S;          % nr. of allowed switches in action sequenvce
dimx = model.p;     % number of state vars

% start counting time
t = cputime;

if ~isempty(n),
	maxtreesize = M*n + 1;
elseif ~isempty(nsim),
	maxtreesize = nsim + 1;
end;
	
% prepare tree data 
Tparent = zeros(1, maxtreesize, 'uint32');  % parent
Tu      = zeros(1, maxtreesize, 'uint8');   % parent action
Tdepth  = zeros(1, maxtreesize, 'uint16');  % depth (uint16 as depth > 65536 considered unfeasible)
Tleaf   = false(1, maxtreesize);            % leaf flag
Tx      = NaN(dimx, maxtreesize, 'double'); % state
Tr      = NaN(1, maxtreesize, 'double');    % reward
TR      = NaN(1, maxtreesize, 'double');    % partial return up to node (cumulative reward)
Ts      = zeros(1, maxtreesize, 'uint8');   % action switches up to current node


% initialize the root node (note 0 = null pointer)
root = 1;
Tparent(root) = 0;
Tu(root) = 0;
Tdepth(root) = 0;
Tleaf(root) = true;
Tx(:,root) = xroot;
Tr(root) = 0;
TR(root) = 0;
Ts(root) = 0;
treesize = 1;

% 2) Main tree construction loop
if ~isempty(n),
	tauMax = n;
elseif ~isempty(nsim),
	tauMax = nsim;
end;
tau=0;
while tau < tauMax, 
    % 2a) select candidate leaves for expansion
    leaves = find((Ts<=S).*Tleaf);
    % optimistic: max b-value RELATIVE TO THE ROOT
    B = TR(leaves) + discount.^double(Tdepth(leaves)) ./ (1-discount);
    leaves = leaves(B + 10^-10 >= max(B));  %!!! bug found for OKP but not applied to OSP till now!!!!!!
    % break any ties between remaining candidates deterministically
    parent = leaves(1);
        
    % 2b) expand the chosen node
    Tleaf(parent) = false;        % no longer a leaf
    
    % run through the discrete actions 
    for j = 1:M,

        % find index where new node will be, save it on the children table of current node
        child = treesize + 1;
        
        % simulate transition and add data to tree
        [Tx(:, child), Tr(child)] = ...
            feval(model.fun, model, Tx(:, parent), U(:, j));
        
%         fprintf('r=%d\n',Tr(child));
        
        % create new node
        Tparent(child) = parent;      % parent is expanded node, along jth action
        Tu(child) = j;      % parent is expanded node, along jth action
        Tdepth(child) = Tdepth(parent) + 1;   % one deeper than the parent
        Tleaf(child) = true;                        % new child is a leaf
        TR(child) = TR(parent) + discount^double(Tdepth(parent)) * Tr(child);
        % if there was a switch in the action sequence, increment counter
        if parent==root || Tu(parent) == Tu(child),
            Ts(child) = Ts(parent);
        else
            Ts(child) = Ts(parent) + 1;
        end        
                    
        % we just added a new node, increase tree size
        treesize = treesize + 1;
    end;
    if ~isempty(nsim),
        tau = tau + M;
    elseif ~isempty(n),
    	tau = tau + 1;
    end;
end;    % end main tree construction loop

% 4) Find action at the tree root
% choose a leaf with maximal u-value (i.e., truncated return)
leaves = find(Tleaf); [Rmax, jmax] = max(TR(leaves)); greedy = leaves(jmax); 
% navigate greedy path upwards in the tree until level 1, just after root
while Tparent(greedy) ~= root, greedy = Tparent(greedy); end;
% choose the action that was taken at root to arrive at this level-1 node
% -- this is the action greedy in util
uroot = U(:, Tu(greedy));
% Note: this avoids the need to store and backup util-values while building the tree

% output stats if requested
if nargout >= 2,
    if ~isempty(n),
    	planinfo.n = tau;
    elseif ~isempty(nsim),
    	planinfo.nsim = tau;
    end;
    planinfo.depth = max(Tdepth(~Tleaf));
    planinfo.trun = cputime - t;
    planinfo.treesize = treesize;
    planinfo.branchfact = 0;
    for xi=1:ceil(max(Tdepth)./3)
        planinfo.branchfact = planinfo.branchfact + sum(Tdepth==planinfo.depth-xi)./sum(Tdepth==planinfo.depth-xi-1);
    end
    planinfo.branchfact = planinfo.branchfact ./ ceil(max(Tdepth)./3);
    planinfo.message = sprintf('Planning done, n=%d expansions, depth=%d, treesize=%d, branching=%.3f', ...
        tau, planinfo.depth, planinfo.treesize, planinfo.branchfact); 
end;
% output only tree data (no memory)
if nargout >= 3,
    plandata.Tparent = Tparent(1:treesize);
    plandata.Tu = Tu(1:treesize);
    plandata.Tdepth = Tdepth(1:treesize);
    plandata.Tleaf = Tleaf(1:treesize);
    plandata.Tx = Tx(:, 1:treesize);
    plandata.Tr = Tr(1:treesize);
    plandata.TR = TR(1:treesize);
    plandata.Ts = Ts(1:treesize);
    plandata.TS = S;    
end;

end % main function