%% You can run this file in cell mode: each cell (starting with %%) can be run by pressing CTRL+ENTER inside it

%% Optimistic Planning for Deterministic systems, OPD

cfg = struct;
% first, configure the problem
cfg.problem = 'ipsetup_problem';
% the reward function is "normalized LQR" = -x' Qrew x - u' Rrew u, but
% scaled and translated into [0,1]
cfg.model_params = {'rewtype=nlqr Qrew=[1,0;0,0.01] Rrew=0.3', 'type=ip maxu=3', 'Ts=0.05'};
% for technical reasons we won't go into, SOPC works well for moderated discount factors
% so for fairness we set the same, smaller discount factor for all algorithms
cfg.gamma = 0.85;
% receding horizon control is implemented generically in GENMPC
% the planning algorithm is changed and configured as follows
cfg.planalg = 'opd';            % use the OPD planner
cfg.planparams.n = 150;         % budget of node expansions at each step
cfg.planparams.U = [-3 0 3];    % discretized action space, in volts for the inv pend
cfg.reset = [-pi, 0]';          % always reset pointing down
cfg.trials = 1;                 % one trial is enough to see performance
cfg.maxsteps = 80;              % 80 steps = 4 seconds should be enough for a swingup
% configure the directory and filename where to save the experiment data
cfg.datadir = pwd;
cfg.datafile = 'ip_opd';
% set to 0 to disable visualization, set to 3 to show a graphical pendulum representation
cfg.visualize = 3;
cfg.viscfg.pause = 0.001;       % shorten the pause to avoid waiting

% run the algorithm
cfg.run = 1;
genmpc(cfg);              
% close visualization figure to avoid them piling up when running multiple experiments
close(gcf);


% look at the controlled trajectory
[opdhist opdfig] = genmpc('plottrial datafile=ip_opd');
% the return is saved on the history, you can obtain it like this
opdr = opdhist.R(1);
% the average execution time across all planning steps is a good representation of the computational
% cost of the algorithm; you can compute it as follows
load ip_opd planinfoh
opdt = mean([planinfoh.trun]);
% when you are done examining you can run following command to close
% the trajectory plot
% close(opdfig);

%% Simultaneous Optimistic Planning with Continuous actions, SOPC

% same problem
cfg = struct;
cfg.problem = 'ipsetup_problem';
cfg.model_params = {'rewtype=nlqr Qrew=[1,0;0,0.01] Rrew=0.3', 'type=ip maxu=3', 'Ts=0.05'};
cfg.gamma = 0.85;
% different planning algorithm
cfg.planalg = 'sooplp';             % continuous-action planner...
cfg.planparams.method = 'SOOP2';    % with SOPC variant
cfg.planparams.dimselect = 'disc';  % dimension selection based on discounting
cfg.planparams.alpha = cfg.gamma;   % with the discount factor same as the MDP
cfg.planparams.n = 400;             % budget of SIMULATIONS (not node expansions) at each step
% note we don't configure any discretized action space; the algorithm is continuous-action
% from here on same config as above
cfg.reset = [-pi, 0]';          % reset pointing down
cfg.trials = 1;                 % one trial is enough to see performance
cfg.maxsteps = 80;              % 80 steps = 4 seconds should be enough for a swingup
cfg.datadir = pwd;              
cfg.datafile = 'ip_sopc';       % best to save to a different file
% set to 0 to disable visualization, set to 3 to show a graphical pendulum representation
cfg.visualize = 3;
cfg.viscfg.pause = 0.001;       % shorten the pause to avoid waiting

% run the algorithm and then close visualization figure
cfg.run = 1;
genmpc(cfg);              
close(gcf);


% controlled trajectories, returns, and execution times are seen in the same way
[sopchist sopcfig] = genmpc('plottrial datafile=ip_sopc');
sopcr = sopchist.R(1);
load ip_sopc planinfoh
sopct = mean([planinfoh.trun]);
% close(sopchist);


